import difflib
from os.path import relpath, join

from pyprint.ConsolePrinter import ConsolePrinter

from coalib.results.Diff import ConflictError
from coalib.results.Result import Result
from coalib.results.result_actions.ResultAction import ResultAction

from coala_utils.decorators import enforce_signature


def format_line(line, real_nr='', sign=']', mod_nr='', symbol='', ):
    return '[{:>4}{}{:1}{}'.format(real_nr,
                                   sign,
                                   symbol,
                                   line.rstrip('\n'))


def print_from_name(printer, line):
    printer.print(format_line(line, real_nr='----'), color='red')


def print_to_name(printer, line):
    printer.print(format_line(line, real_nr='++++'), color='green')


def print_beautified_diff(difflines, printer):
    current_line_added = None
    current_line_subtracted = None
    for line in difflines:
        if line.startswith('@@'):
            values = line[line.find('-'):line.rfind(' ')]
            subtracted, added = tuple(values.split(' '))
            current_line_added = int(added.split(',')[0][1:])
            current_line_subtracted = int(subtracted.split(',')[0][1:])
        elif line.startswith('---'):
            print_from_name(printer, line[4:])
        elif line.startswith('+++'):
            print_to_name(printer, line[4:])
        elif line.startswith('+'):
            printer.print(format_line(line[1:], real_nr=current_line_added),
                          color='green')
            current_line_added += 1
        elif line.startswith('-'):
            printer.print(format_line(line[1:],
                                      real_nr=current_line_subtracted),
                          color='red')
            current_line_subtracted += 1
        else:
            current_line_subtracted += 1
            current_line_added += 1


class ShowPatchAction(ResultAction):

    SUCCESS_MESSAGE = 'Displayed patch successfully.'

    @staticmethod
    @enforce_signature
    def is_applicable(result: Result, original_file_dict, file_diff_dict):

        if not result.diffs:
            return 'This result has no patch attached.'

        try:
            # Needed so the addition is run for all patches -> ConflictError
            nonempty_patches = False
            for filename, diff in result.diffs.items():
                if diff and (filename not in file_diff_dict or
                             diff + file_diff_dict[filename] !=
                             file_diff_dict[filename]):
                    nonempty_patches = True

            if nonempty_patches:
                return True
            return 'The given patches do not change anything anymore.'

        except ConflictError as ce:
            return ('Two or more patches conflict with '
                    'each other: {}'.format(str(ce)))

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict,
              no_color: bool=False,
              show_result_on_top: bool=False):
        """
        (S)how patch

        :param no_color:
            Whether or not to use colored output.
        :param show_result_on_top:
            Set this to True if you want to show the result info on top.
            (Useful for e.g. coala_ci.)
        """
        printer = ConsolePrinter(not no_color)

        if show_result_on_top:
            from coalib.output.ConsoleInteraction import print_result
            # Most of the params are empty because they're unneeded in
            # noninteractive mode. Yes, this cries for a refactoring...
            print_result(printer, None, {}, result, {}, interactive=False)

        for filename, this_diff in sorted(result.diffs.items()):
            to_filename = this_diff.rename if this_diff.rename else filename
            to_filename = '/dev/null' if this_diff.delete else to_filename
            original_file = original_file_dict[filename]
            try:
                current_file = file_diff_dict[filename].modified
                new_file = (file_diff_dict[filename] + this_diff).modified
            except KeyError:
                current_file = original_file
                new_file = this_diff.modified

            if tuple(current_file) != tuple(new_file):
                print_beautified_diff(difflib.unified_diff(current_file,
                                                           new_file,
                                                           fromfile=filename,
                                                           tofile=to_filename),
                                      printer)
            elif filename != to_filename:
                print_from_name(printer, join('a', relpath(filename)))
                print_to_name(printer, join('b', relpath(to_filename)))

        return file_diff_dict



from coalib.parsing.filters.Filter import filter


@filter
def can_fix_filter(bear, args):
    """
    Filters the bears by ``CAN_FIX``.

    :param bear: Bear object.
    :param args: Set of fixable issue types on which ``bear`` is to be
                 filtered.
    :return:     ``True`` if this bear matches the criteria inside args,
                 ``False`` otherwise.
    """
    return bool({fix.lower() for fix in bear.CAN_FIX} & args)



from collections import Iterable, namedtuple
from glob import iglob
import os.path

from coala_utils.decorators import (
    enforce_signature, generate_eq, generate_repr)
from coalib.parsing.ConfParser import ConfParser


@generate_repr()
@generate_eq('language', 'docstyle', 'markers')
class DocstyleDefinition:
    """
    The DocstyleDefinition class holds values that identify a certain type of
    documentation comment (for which language, documentation style/tool used
    etc.).
    """
    Metadata = namedtuple('Metadata', ('param_start', 'param_end',
                                       'exception_start', 'exception_end',
                                       'return_sep'))
    ClassPadding = namedtuple('ClassPadding',
                              ('top_padding', 'bottom_padding'))
    FunctionPadding = namedtuple('FunctionPadding',
                                 ('top_padding', 'bottom_padding'))
    DocstringTypeRegex = namedtuple('DocstringTypeRegex',
                                    ('class_sign', 'func_sign'))

    @enforce_signature
    def __init__(self, language: str, docstyle: str, markers: (Iterable, str),
                 metadata: Metadata, class_padding: ClassPadding,
                 function_padding: FunctionPadding,
                 docstring_type_regex: DocstringTypeRegex,
                 docstring_position: str):
        """
        Instantiates a new DocstyleDefinition.

        :param language: The case insensitive programming language of the
                         documentation comment, e.g. ``"CPP"`` for C++ or
                         ``"PYTHON3"``.
        :param docstyle: The case insensitive documentation style/tool used
                         to document code, e.g. ``"default"`` or ``"doxygen"``.
        :param markers:  An iterable of marker/delimiter string iterables
                         or a single marker/delimiter string iterable that
                         identify a documentation comment. See ``markers``
                         property for more details on markers.
        :param metadata: A namedtuple consisting of certain attributes that
                         form the layout of the certain documentation comment
                         e.g. ``param_start`` defining the start symbol of
                         the parameter fields and ``param_end`` defining the
                         end.
        :param class_padding: A namedtuple consisting of values about
                         blank lines before and after the documentation of
                         ``docstring_type`` class.
        :param function_padding: A namedtuple consisting of values about
                         blank lines before and after the documentation of
                         ``docstring_type`` function.
        :param docstring_type_regex: A namedtuple consisting of regex
                         about ``class`` and ``function`` of a language, which
                         is used to determine ``docstring_type`` of
                         DocumentationComment.
        :param docstring_position: Defines the position where the regex of
                         docstring type is present(i.e. ``top`` or ``bottom``).
        """
        self._language = language.lower()
        self._docstyle = docstyle.lower()

        # Check and modify tuple if only one marker_set exists.
        markers = tuple(markers)
        if len(markers) == 3 and all(isinstance(x, str) for x in markers):
            markers = (markers,)

        self._markers = tuple(tuple(marker_set) for marker_set in markers)

        # Check marker set dimensions.
        for marker_set in self._markers:
            length = len(marker_set)
            if length != 3:
                raise ValueError('Length of a given marker set was not 3 (was '
                                 'actually {}).'.format(length))

        self._metadata = metadata
        self._class_padding = class_padding
        self._function_padding = function_padding
        self._docstring_type_regex = docstring_type_regex
        self._docstring_position = docstring_position

    @property
    def language(self):
        """
        The programming language.

        :return: A lower-case string defining the programming language (i.e.
                 "cpp" or "python").
        """
        return self._language

    @property
    def docstyle(self):
        """
        The documentation style/tool used to document code.

        :return: A lower-case string defining the docstyle (i.e. "default" or
                 "doxygen").
        """
        return self._docstyle

    @property
    def markers(self):
        """
        A tuple of marker sets that identify a documentation comment.

        Marker sets consist of 3 entries where the first is the start-marker,
        the second one the each-line marker and the last one the end-marker.
        For example a marker tuple with a single marker set
        ``(("/**", "*", "*/"),)`` would match following documentation comment:

        ::

            /**
             * This is documentation.
             */

        It's also possible to supply an empty each-line marker
        (``("/**", "", "*/")``):

        ::

            /**
             This is more documentation.
             */

        Markers are matched "greedy", that means it will match as many
        each-line markers as possible. I.e. for ``("///", "///", "///")``):

        ::

            /// Brief documentation.
            ///
            /// Detailed documentation.

        :return: A tuple of marker/delimiter string tuples that identify a
                 documentation comment.
        """
        return self._markers

    @property
    def metadata(self):
        """
        A namedtuple of certain attributes present in the documentation.

        These attributes are used to define parts of the documentation.
        """
        return self._metadata

    @property
    def class_padding(self):
        """
        A namedtuple ``ClassPadding`` consisting of values about blank lines
        before and after the documentation of ``docstring_type`` class.

        These values are official standard of following blank lines before and
        after the documentation of ``docstring_type`` class.
        """
        return self._class_padding

    @property
    def function_padding(self):
        """
        A namedtuple ``FunctionPadding`` consisting of values about blank
        lines before and after the documentation of ``docstring_type``
        function.

        These values are official standard of following blank lines before and
        after the documentation of ``docstring_type`` function.
        """
        return self._function_padding

    @property
    def docstring_type_regex(self):
        """
        A namedtuple ``DocstringTypeRegex`` consisting of regex about ``class``
        and ``function`` of a language, which is used to determine
        ``docstring_type`` of DocumentationComment.
        """
        return self._docstring_type_regex

    @property
    def docstring_position(self):
        """
        Defines the position, where the regex of docstring type is present.
        Depending on different languages the docstrings are present below or
        above the defined class or function. This expicitly defines where the
        class regex or function regex is present(i.e. ``top`` or ``bottom``).
        """
        return self._docstring_position

    @classmethod
    @enforce_signature
    def load(cls, language: str, docstyle: str, coalang_dir=None):
        """
        Loads a ``DocstyleDefinition`` from the coala docstyle definition files.

        This function considers all settings inside the according coalang-files
        as markers, except ``param_start``, ``param_end`` and ``return_sep``
        which are considered as special metadata markers.

        .. note::

            When placing new coala docstyle definition files, these must
            consist of only lowercase letters and end with ``.coalang``!

        :param language:           The case insensitive programming language of
                                   the documentation comment as a string.
        :param docstyle:           The case insensitive documentation
                                   style/tool used to document code, e.g.
                                   ``"default"`` or ``"doxygen"``.
        :param coalang_dir:        Path to directory with coalang docstyle
                                   definition files. This replaces the default
                                   path if given.
        :raises FileNotFoundError: Raised when the given docstyle was not
                                   found.
        :raises KeyError:          Raised when the given language is not
                                   defined for given docstyle.
        :return:                   The ``DocstyleDefinition`` for given language
                                   and docstyle.
        """

        docstyle = docstyle.lower()

        language_config_parser = ConfParser(remove_empty_iter_elements=False)

        coalang_file = os.path.join(
            coalang_dir or os.path.dirname(__file__), docstyle + '.coalang')

        try:
            docstyle_settings = language_config_parser.parse(coalang_file)
        except FileNotFoundError:
            raise FileNotFoundError('Docstyle definition ' + repr(docstyle) +
                                    ' not found.')

        language = language.lower()

        try:
            docstyle_settings = docstyle_settings[language]
        except KeyError:
            raise KeyError('Language {!r} is not defined for docstyle {!r}.'
                           .format(language, docstyle))

        metadata_settings = ('param_start', 'param_end',
                             'exception_start', 'exception_end',
                             'return_sep')

        metadata = cls.Metadata(*(str(docstyle_settings.get(req_setting, ''))
                                  for req_setting in metadata_settings))

        try:
            class_padding = cls.ClassPadding(
                *(int(padding) for padding in tuple(
                    docstyle_settings['class_padding'])))
        except IndexError:
            class_padding = cls.ClassPadding('', '')

        try:
            function_padding = cls.FunctionPadding(
                *(int(padding) for padding in tuple(
                    docstyle_settings['function_padding'])))
        except IndexError:
            function_padding = cls.FunctionPadding('', '')

        try:
            docstring_type_regex = cls.DocstringTypeRegex(
                *(str(sign) for sign in tuple(
                    docstyle_settings['docstring_type_regex'])))
        except IndexError:
            docstring_type_regex = cls.DocstringTypeRegex('', '')

        try:
            docstring_position = docstyle_settings['docstring_position'].value
        except IndexError:
            docstring_position = ''

        ignore_keys = (('class_padding', 'function_padding',
                        'docstring_type_regex', 'docstring_position') +
                       metadata_settings)

        marker_sets = (tuple(value)
                       for key, value in
                       docstyle_settings.contents.items()
                       if key not in ignore_keys and
                       not key.startswith('comment'))

        return cls(language, docstyle, marker_sets, metadata, class_padding,
                   function_padding, docstring_type_regex, docstring_position)

    @staticmethod
    def get_available_definitions():
        """
        Returns a sequence of pairs with ``(docstyle, language)`` which are
        available when using ``load()``.

        :return: A sequence of pairs with ``(docstyle, language)``.
        """
        pattern = os.path.join(os.path.dirname(__file__), '*.coalang')

        for coalang_file in iglob(pattern):
            docstyle = os.path.splitext(os.path.basename(coalang_file))[0]
            # Ignore files that are not lowercase, as coalang files have to be.
            if docstyle.lower() == docstyle:
                parser = ConfParser(remove_empty_iter_elements=False)
                for language in parser.parse(coalang_file):
                    yield docstyle, language.lower()



from coalib.results.result_actions.ResultAction import ResultAction


class DoNothingAction(ResultAction):

    SUCCESS_MESSAGE = ''

    is_applicable = staticmethod(lambda *args: True)

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict):
        """
        Do (N)othing
        """



from coalib.parsing.InvalidFilterException import InvalidFilterException

from coalib.parsing.filters.LanguageFilter import language_filter
from coalib.parsing.filters.CanDetectFilter import can_detect_filter
from coalib.parsing.filters.CanFixFilter import can_fix_filter


available_filters = {'language': language_filter,
                     'can_detect': can_detect_filter,
                     'can_fix': can_fix_filter}


def is_valid_filter(filter):
    return filter in available_filters


def apply_filter(filter_name, filter_args, all_bears=None):
    if all_bears is None:
        from coalib.settings.ConfigurationGathering import (
            get_all_bears)
        all_bears = get_all_bears()
    if not is_valid_filter(filter_name):
        raise InvalidFilterException('{!r} is an invalid filter. '
                                     'Available filters: {}'.format(
                                         filter_name,
                                         ', '.join(sorted(
                                             available_filters))))
    if not filter_args or len(filter_args) == 0:
        return all_bears
    return available_filters[filter_name](all_bears, filter_args)


def apply_filters(filters, bears=None):
    """
    Returns bears after filtering based on ``args``. It returns
    intersection of bears if more than one element is present in ``args``
    list.

    :param filters:
        List of args based on ``bears`` has to be filtered.
        For example:
        ``[['language', 'c', 'java'], ['can_fix', 'syntax']]``
    :param bears:
        The bears to filter.
    :return:
        Filtered bears.
    """
    for filter in filters:
        filter_name, *filter_args = filter
        bears = apply_filter(filter_name, filter_args, bears)
    return bears



import itertools

import coalib.bearlib.aspects
from coalib.bearlib.aspects.meta import isaspect, issubaspect
from coalib.bearlib.aspects.base import aspectbase


class AspectList(list):
    """
    List-derived container to hold aspects.
    """

    def __init__(self, seq=(), exclude=None, languages=None):
        """
        Initialize a new AspectList.

        >>> from .Metadata import CommitMessage
        >>> AspectList([CommitMessage.Shortlog, CommitMessage.Body])
        [<aspectclass '...Shortlog'>, <aspectclass '...Body'>]
        >>> AspectList(['Shortlog', 'CommitMessage.Body'])
        [<aspectclass '...Shortlog'>, <aspectclass '...Body'>]
        >>> AspectList([CommitMessage.Shortlog, 'CommitMessage.Body'])
        [<aspectclass '...Shortlog'>, <aspectclass '...Body'>]

        :param seq: A sequence containing either aspectclass, aspectclass
                    instance, or string of partial/full qualified aspect name.
        :param exclude: A sequence of either aspectclass or string of aspect
                        name that marked as excluded from the list.
        """
        super().__init__((item if isaspect(item) else
                          coalib.bearlib.aspects[item] for item in seq))

        self.languages = languages
        self.exclude = AspectList(exclude) if exclude is not None else []

    def __contains__(self, aspect):
        # Check if ``aspects`` language is supported.
        if self.languages is not None and isinstance(aspect, aspectbase):
            if aspect.language not in self.languages:
                return False

        for item in self:
            if issubaspect(aspect, item):
                return aspect not in self.exclude
        return False

    def get(self, aspect):
        """
        Return first item that match or contain an aspect. See
        :meth:`coalib.bearlib.aspects.aspectbase.get` for further example.

        :param aspect: An aspectclass OR name of an aspect.
        :return:       An aspectclass OR aspectclass instance, depend on
                       AspectList content. Return None if no match found.
        """
        if not isaspect(aspect):
            aspect = coalib.bearlib.aspects[aspect]
        if aspect in self.exclude:
            return None
        try:
            return next(filter(None, (item.get(aspect) for item in self)))
        except StopIteration:
            return None

    def _remove(self, item):
        """
        Remove first matching item in list.

        :param item:        An aspectclass
        :raises ValueError: When to be removed item is not found in list.
        """
        for aspect in self:
            if aspect is item or isinstance(aspect, item):
                return super().remove(aspect)

        raise ValueError('{}._remove(x): {} not in list.'
                         .format(type(self).__name__, item))

    def get_leaf_aspects(self):
        """
        Breakdown all of item in self into their leaf subaspects.

        :return: An AspectList contain ONLY leaf aspects.
        """
        aspects = type(self)()
        for leaf_aspect in itertools.chain.from_iterable([
                aspect.get_leaf_aspects() for aspect in self]):
            # Make sure no duplication
            if leaf_aspect not in aspects:
                aspects.append(leaf_aspect)

        for excluded_aspect in itertools.chain.from_iterable(
                [aspect.get_leaf_aspects() for aspect in self.exclude]):
            try:
                aspects._remove(excluded_aspect)
            except ValueError:
                continue

        return aspects



"""
Language and docstyle independent extraction of documenation comments.

Each of the functions is built upon one another, and at the last,
exposes a single function :func:`extract_documentation_with_markers`
which is used by :class:`.DocBaseClass`, to extract documentation.
"""

import re

from coalib.bearlib.languages.documentation.DocumentationComment import (
    DocumentationComment, MalformedComment)
from coalib.results.TextPosition import TextPosition
from coalib.results.TextRange import TextRange
from textwrap import dedent


def _extract_doc_comment_simple(content, line, column, markers):
    """
    Extract a documentation that starts at given beginning with simple layout.

    The property of the simple layout is that there's no each-line marker. This
    applies e.g. for python docstrings.

    :param content: Presplitted lines of the source-code-string.
    :param line:    Line where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param column:  Column where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param markers: The documentation identifying markers.
    :return:        If the comment matched layout a triple with end-of-comment
                    line, column and the extracted documentation. If not
                    matched, returns None.
    """
    align_column = column - len(markers[0])

    pos = content[line].find(markers[2], column)
    if pos != -1:
        return line, pos + len(markers[2]), content[line][column:pos]

    doc_comment = content[line][column:]
    line += 1

    while line < len(content):
        pos = content[line].find(markers[2])
        if pos == -1:
            line_column = len(content[line])-len(content[line].lstrip())
            doc_comment += ('\n' if content[line][align_column:] == ''
                            else content[line].strip()+'\n'
                            if line_column < align_column
                            else content[line][align_column:])
        else:
            doc_comment += content[line][align_column:pos]
            return line, pos + len(markers[2]), doc_comment

        line += 1

    return None


def _extract_doc_comment_continuous(content, line, column, markers):
    """
    Extract a documentation that starts at given beginning with continuous
    layout.

    The property of the continuous layout is that the each-line-marker and the
    end-marker do equal. Documentation is extracted until no further marker is
    found. Applies e.g. for doxygen style python documentation::

        ## main
        #
        #  detailed

    :param content: Presplitted lines of the source-code-string.
    :param line:    Line where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param column:  Column where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param markers: The documentation identifying markers.
    :return:        If the comment matched layout a triple with end-of-comment
                    line, column and the extracted documentation. If not
                    matched, returns None.
    """
    marker_len = len(markers[1])

    doc_comment = content[line][column:]
    line += 1
    while line < len(content):
        pos = content[line].find(markers[1])
        if pos == -1:
            return line, 0, doc_comment
        else:
            doc_comment += content[line][pos + marker_len:]

        line += 1

    if content[line - 1][-1] == '\n':
        column = 0
    else:
        # This case can appear on end-of-document without a ``\n``.
        line -= 1
        column = len(content[line])

    return line, column, doc_comment


def _extract_doc_comment_standard(content, line, column, markers):
    """
    Extract a documentation that starts at given beginning with standard
    layout.

    The standard layout applies e.g. for C doxygen-style documentation::

        /**
         * documentation
         */

    :param content: Presplitted lines of the source-code-string.
    :param line:    Line where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param column:  Column where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param markers: The documentation identifying markers.
    :return:        If the comment matched layout a triple with end-of-comment
                    line, column and the extracted documentation. If not
                    matched, returns None.
    """
    pos = content[line].find(markers[2], column)
    if pos != -1:
        return line, pos + len(markers[2]), content[line][column:pos]

    doc_comment = content[line][column:]
    line += 1

    while line < len(content):
        pos = content[line].find(markers[2])
        each_line_pos = content[line].find(markers[1])

        if pos == -1:
            if each_line_pos == -1:
                # If the first text occurrence is not the each-line marker
                # now we violate the doc-comment layout.
                return None
            doc_comment += content[line][each_line_pos + len(markers[1]):]
        else:
            # If no each-line marker found or it's located past the end marker:
            # extract no further and end the doc-comment.
            if each_line_pos != -1 and each_line_pos + 1 < pos:
                doc_comment += content[line][each_line_pos +
                                             len(markers[1]):pos]

            return line, pos + len(markers[2]), doc_comment

        line += 1

    return None


def _extract_doc_comment(content, line, column, markers):
    """
    Delegates depending on the given markers to the right extraction method.

    :param content: Presplitted lines of the source-code-string.
    :param line:    Line where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param column:  Column where the documentation comment starts (behind the
                    start marker). Zero-based.
    :param markers: The documentation identifying markers.
    :return:        If the comment matched layout a triple with end-of-comment
                    line, column and the extracted documentation. If not
                    matched, returns None.
    """
    if markers[1] == '':
        # Extract and align to start marker.
        return _extract_doc_comment_simple(content, line, column, markers)
    elif markers[1] == markers[2]:
        # Search for the each-line marker until it runs out.
        return _extract_doc_comment_continuous(content, line, column, markers)
    else:
        return _extract_doc_comment_standard(content, line, column, markers)


def _compile_multi_match_regex(strings):
    """
    Compiles a regex object that matches each of the given strings.

    :param strings: The strings to match.
    :return:        A regex object.
    """
    return re.compile('|'.join(re.escape(s) for s in strings))


def _extract_doc_comment_from_line(content, line, column, regex,
                                   marker_dict, docstyle_definition):
    cur_line = content[line]
    begin_match = regex.search(cur_line, column)
    if begin_match:
        indent = cur_line[:begin_match.start()]
        column = begin_match.end()
        for marker in marker_dict[begin_match.group()]:
            doc_comment = _extract_doc_comment(content, line, column, marker)
            if doc_comment is not None:
                end_line, end_column, documentation = doc_comment

                position = TextPosition(line + 1, len(indent) + 1)
                doc = DocumentationComment(documentation, docstyle_definition,
                                           indent, marker, position)

                break

        if doc_comment:
            return end_line, end_column, doc
        else:
            malformed_comment = MalformedComment(dedent("""\
                Please check the docstring for faulty markers. A starting
                marker has been found, but no instance of DocComment is
                returned."""), line)
            return line + 1, 0, malformed_comment

    return line + 1, 0, None


def extract_documentation_with_markers(content, docstyle_definition):
    """
    Extracts all documentation texts inside the given source-code-string.

    :param content:
        The source-code-string where to extract documentation from.
        Needs to be a list or tuple where each string item is a single
        line (including ending whitespaces like ``\\n``).
    :param docstyle_definition:
        The ``DocstyleDefinition`` instance that defines what docstyle is
        being used in the documentation.
    :return:
        An iterator returning each DocumentationComment found in the content.
    """
    # Prepare marker-tuple dict that maps a begin pattern to the corresponding
    # marker_set(s). This makes it faster to retrieve a marker-set from a
    # begin sequence we initially want to search for in source code. Then
    # the possible found documentation match is processed further with the
    # rest markers.
    markers = docstyle_definition.markers

    marker_dict = {}
    for marker_set in markers:
        if marker_set[0] not in marker_dict:
            marker_dict[marker_set[0]] = [marker_set]
        else:
            marker_dict[marker_set[0]].append(marker_set)

    # Using regexes to perform a variable match is faster than finding each
    # substring with ``str.find()`` choosing the lowest match.
    begin_regex = _compile_multi_match_regex(
        marker_set[0] for marker_set in markers)

    line = 0
    column = 0
    while line < len(content):
        line, column, doc = _extract_doc_comment_from_line(
            content,
            line,
            column,
            begin_regex,
            marker_dict,
            docstyle_definition)

        if doc and isinstance(doc, MalformedComment):
            yield doc
        elif doc:
            # Ignore string literals
            ignore_regex = re.compile(
                '^\s*r?(?P<marker>' +
                ('|'.join(re.escape(s) for s in doc.marker[0])) +
                ')')
            # Starting line of doc_string where marker is present
            start_line = doc.range.start.line - 1
            ignore_string_match = ignore_regex.search(content[start_line])

            # Instantiate padding
            top_padding = 0
            bottom_padding = 0
            # minus 2 because we want to check the line before the marker.
            start_index = doc.range.start.line - 2
            end_index = doc.range.end.line
            while start_index >= 0 and not content[start_index].strip():
                top_padding += 1
                start_index -= 1
            # If the end_index is instantiated above the len(content) i.e.
            # In case where ending marker of docstring is at the last line.
            # Then the doc.bottom_padding will be default to 0. This will also
            # prevent IndexError raised by content[end_index].
            while end_index < len(content) and not content[end_index].strip():
                # This condition will take place if theres an inline docstring
                # following documentation.
                if ((doc.marker[2]+'\n') != content[end_index-1][-4:]
                        and bottom_padding == 0):
                    break
                bottom_padding += 1
                end_index += 1

            class_regex = re.compile(
                doc.docstyle_definition.docstring_type_regex.class_sign)
            function_regex = re.compile(
                doc.docstyle_definition.docstring_type_regex.func_sign)

            # End line differs when mid marker and end marker is different
            if doc.marker[1] == doc.marker[2]:
                end_index = end_index - 1

            # Check for docstring_position and then check for class regex
            # and function regex to define the type of docstring.
            if doc.docstyle_definition.docstring_position == 'top':
                if class_regex.search(content[start_index]):
                    doc.docstring_type = 'class'
                elif function_regex.search(content[start_index]):
                    doc.docstring_type = 'function'
            elif doc.docstyle_definition.docstring_position == 'bottom':
                if (end_index < len(content) and
                        class_regex.search(content[end_index])):
                    doc.docstring_type = 'class'
                elif (end_index < len(content) and
                        function_regex.search(content[end_index])):
                    doc.docstring_type = 'function'

            # Disabled automatic padding for docstring_type='others' as this
            # will cause overlapping of range in consecutive docstrings. Which
            # diff.replace() is unable to handle.
            if doc.docstring_type != 'others':
                doc.top_padding = top_padding
                doc.bottom_padding = bottom_padding

                doc.range = TextRange.from_values(
                    start_index + 2,
                    1 if top_padding > 0 else doc.range.start.column,
                    end_index,
                    1 if bottom_padding > 0 else doc.range.end.column)

            if ignore_string_match:
                yield doc



import re

from inspect import getmembers, signature

from coala_utils.decorators import generate_repr

from .base import aspectbase
from .docs import Documentation
from .exceptions import AspectTypeError
from .taste import Taste


class aspectclass(type):
    """
    Metaclass for aspectclasses.

    Root aspectclass is :class:`coalib.bearlib.aspects.Root`.
    """
    def __init__(cls, clsname, bases, clsattrs):
        """
        Initializes the ``.subaspects`` dict on new aspectclasses.
        """
        cls.subaspects = {}

    @property
    def tastes(cls):
        """
        Get a dictionary of all taste names mapped to their
        :class:`coalib.bearlib.aspects.Taste` instances.
        """
        if cls.parent:
            return dict(cls.parent.tastes, **cls._tastes)

        return dict(cls._tastes)

    def subaspect(cls, subcls):
        """
        The sub-aspectclass decorator.

        See :class:`coalib.bearlib.aspects.Root` for description
        and usage.
        """
        aspectname = subcls.__name__
        sub_qualname = '%s.%s' % (cls.__qualname__, aspectname)

        docs = getattr(subcls, 'docs', None)
        aspectdocs = Documentation(subcls.__doc__, **{
            attr: getattr(docs, attr, '') for attr in
            list(signature(Documentation).parameters.keys())[1:]})

        # search for tastes in the sub-aspectclass
        subtastes = {}
        for name, member in getmembers(subcls):
            if isinstance(member, Taste):
                # tell the taste its own name
                member.name = name
                # tell its owner name
                member.aspect_name = sub_qualname
                subtastes[name] = member

        class Sub(subcls, aspectbase, metaclass=aspectclass):
            __module__ = subcls.__module__

            parent = cls

            docs = aspectdocs
            _tastes = subtastes

        members = sorted(Sub.tastes)
        if members:
            Sub = generate_repr(*members)(Sub)

        Sub.__name__ = aspectname
        Sub.__qualname__ = sub_qualname
        cls.subaspects[aspectname] = Sub
        setattr(cls, aspectname, Sub)
        return Sub

    def __repr__(cls):
        return '<%s %s>' % (type(cls).__name__, repr(cls.__qualname__))


def isaspect(item):
    """
    This function checks whether or not an object is an ``aspectclass`` or an
    instance of ``aspectclass``
    """
    return isinstance(item, (aspectclass, aspectbase))


def assert_aspect(item):
    """
    This function raises ``AspectTypeError`` when an object is not an
    ``aspectclass`` or an instance of ``aspectclass``
    """
    if not isaspect(item):
        raise AspectTypeError(item)
    return item


def issubaspect(subaspect, aspect):
    """
    This function checks whether or not ``subaspect`` is a subaspect of
    ``aspect``.
    """
    subaspect = assert_aspect(subaspect)
    aspect = assert_aspect(aspect)
    aspect_qualname = (aspect.__qualname__ if isinstance(
                    aspect, aspectclass) else type(aspect).__qualname__)
    subaspect_qualname = (subaspect.__qualname__ if isinstance(
                    subaspect, aspectclass) else type(subaspect).__qualname__)
    return re.match(aspect_qualname+'(\.|$)', subaspect_qualname) is not None



import logging


def fail_acquire_settings(log_printer, settings_names_dict, section):
    """
    This method throws an exception if any setting needs to be acquired.

    :param log_printer:         Printer responsible for logging the messages.
    :param settings_names_dict: A dictionary with the settings name as key and
                                a list containing a description in [0] and the
                                name of the bears who need this setting in [1]
                                and following.
    :raises AssertionError:     If any setting is required.
    :raises TypeError:          If ``settings_names_dict`` is not a
                                dictionary.
    """
    if not isinstance(settings_names_dict, dict):
        raise TypeError('The settings_names_dict parameter has to be a '
                        'dictionary.')

    required_settings = settings_names_dict.keys()
    if len(required_settings) != 0:
        msg = ('During execution, we found that some required '
               'settings were not provided. They are:\n')

        for name, setting in settings_names_dict.items():
            msg += '{} (from {}) - {}'.format(name, setting[1], setting[0])

        logging.error(msg)
        raise AssertionError(msg)



import logging
import queue
import threading

from coalib.processes.communication.LogMessage import LogMessage


class LogPrinterThread(threading.Thread):
    """
    This is the Thread object that outputs all log messages it gets from
    its message_queue. Setting obj.running = False will stop within the next
    0.1 seconds.
    """

    def __init__(self, message_queue, log_printer=None):
        threading.Thread.__init__(self)
        self.running = True
        self.message_queue = message_queue

    def run(self):
        while self.running:
            try:
                elem = self.message_queue.get(timeout=0.1)
                if isinstance(elem, LogMessage):
                    logging.log(elem.log_level, elem.message)
                else:
                    logging.info(elem)
            except queue.Empty:
                pass



from coalib.bearlib.languages import Language
from coalib.bearlib.aspects import Root, Taste


@Root.subaspect
class Redundancy:
    """
    This aspect describes redundancy in your source code.
    """
    class docs:
        example = """
        int foo(int iX)
        {
            int iY = iX*2;

            return iX*2;
        }
        """
        example_language = 'C++'
        importance_reason = """
        Redundant code makes your code harder to read and understand.
        """
        fix_suggestions = """
        Redundant code can usually be removed without consequences.
        """


@Redundancy.subaspect
class Clone:
    """
    Code clones are multiple pieces of source code in your
    codebase that are very similar.
    """
    class docs:
        example = """
        extern int array_a[];
        extern int array_b[];

        int sum_a = 0;

        for (int i = 0; i < 4; i++)
            sum_a += array_a[i];

        int average_a = sum_a / 4;

        int sum_b = 0;

        for (int i = 0; i < 4; i++)
            sum_b += array_b[i];

        int average_b = sum_b / 4;
        """
        example_language = 'C++'
        importance_reason = """
        Code clones make editing more difficult due to unnecessary increases
        in complexity and length.
        """
        fix_suggestions = """
        Usually code clones can be simplified to only one occurrence. In a
        lot of cases, both just use different values or variables and can
        be reduced to one function called with different parameters or
        loops.
        """

    min_clone_tokens = Taste[int](
        'The number of tokens that have to be equal for it to'
        ' be detected as a code clone.',
        (20, ), default=20)

    ignore_using = Taste[bool](
        'Ignore ``using`` directives in C#.',
        (True, False), default=False,
        languages=(Language.CSharp, ))


@Redundancy.subaspect
class UnusedImport:
    """
    Unused imports are any kind of import/include that is not needed.

    This aspect have following taste:

    >>> len(UnusedImport.tastes)
    1
    >>> UnusedImport.remove_non_standard_import
    <...Taste[bool] object at 0x...>
    >>> UnusedImport.remove_non_standard_import.default
    True
    """
    class docs:
        example = """
        import sys
        import os

        print('coala is always written with lowercase c')
        """
        example_language = 'python'
        importance_reason = """
        Redundant imports can cause a performance degrade and make code
        harder to understand when reading through it. Also it causes
        unneeded dependencies within your modules. In most programming
        languages, unused imports may have side effects and that may
        be a common false positive. However those should be avoided.
        """
        fix_suggestions = """
        Usually, unused imports can simply be removed.
        """

    remove_non_standard_import = Taste[bool](
        "Remove ALL unused import, include those not from language's "
        'standard library.',
        (True, False), default=True)


@Redundancy.subaspect
class UnreachableCode:
    """
    Unreachable code, sometimes called dead code, is source code that
    can never be executed during the program execution.
    """
    class docs:
        example = """
        def func():
            return True

        if func():
            a = {}
        else:
            a = (i for i in range (5))
            print (id(a))
        """
        example_language = 'python'
        importance_reason = """
        Unreachable code, makes the source code longer and more difficult
        to maintain.
        """
        fix_suggestions = """
        Those pieces of code can easily be removed without consequences.
        """


@UnreachableCode.subaspect
class UnusedFunction:
    """
    An unused function is a function that is never called during
    code execution.
    """
    class docs:
        example = """
        def func():
            pass

        print('coala is always written with lowercase c')
        """
        example_language = 'python'
        importance_reason = """
        Unused functions make the source code more longer and more
        difficult to maintain.
        """
        fix_suggestions = """
        It is recommended to remove those functions. If you would like
        to access it's source code later for other purposes, you can
        rely on a version control system like Git, Mercurial or
        Subversion.
        """


@UnreachableCode.subaspect
class UnreachableStatement:
    """
    An unreachable statement is a statement that is never executed
    during code execution.
    """
    class docs:
        example = """
        def func():
            return True

        if func():
            a = {}
        else:
            a = (i for i in range (5))
            print (id(a))
        """
        example_language = 'python'
        importance_reason = """
        We should always keep our codebase clean, having dead code uselessly
        makes the code longer and ambiguous.
        """
        fix_suggestions = """
        These statement can be remove without harming the code.
        """


@Redundancy.subaspect
class UnusedVariable:
    """
    Unused variables are declared but never used.
    """
    class docs:
        example = """
        a = {}
        print ('coala')
        """
        example_language = 'python'
        importance_reason = """
        Unused variables can degrade performance marginally but more importantly
        makes the source code harder to read and understand.
        """
        fix_suggestions = """
        Those variables can easily be removed without consequences.
        """


@UnusedVariable.subaspect
class UnusedParameter:
    """
    Unused parameters are functions arguments which are never used.
    """
    class docs:
        example = """
        def func(a):
            pass
        """
        example_language = 'python'
        importance_reason = """
        Unused paramaters are useless to functions, they them difficult to
        use and maintain.
        """
        fix_suggestions = """
        Those parameters can easily be removed without consequences.
        """


@UnusedVariable.subaspect
class UnusedLocalVariable:
    """
    These are variable which are defined locally but never used.
    """
    class docs:
        example = """
        def func():
            for i in range (5):
                a = 0
                print ( ' coala ' )
        """
        example_language = 'python'
        importance_reason = """
        They make the code difficult to maintain.
        """
        fix_suggestions = """
        These can easily be removed without consequences.
        """


@UnusedVariable.subaspect
class UnusedGlobalVariable:
    """
    These are variable which have a global scope but are never used.
    """
    class docs:
        example = """
        a = 0
        for i in range (5):
            print ( ' coala ' )
        """
        example_language = 'python'
        importance_reason = """
        They make the code difficult to maintain.
        """
        fix_suggestions = """
        These can easily be removed without consequences.
        """
